package pascal.p4code;

import pascal.analysis.*;
import pascal.node.*;
import pascal.table.*;
import pascal.iterator.*;

import java.io.*;
import java.util.Hashtable;
import java.util.LinkedList;


public class CodeGenerator extends DepthFirstAdapter{
  SymbolTable table;
  Hashtable info;
  Code c;
  Label l;

  Type type_int, type_real, type_char, type_bool, type_string;

  final int null_val=0;
  final double Null_Val=0.0;


  public CodeGenerator() throws Exception{
    TableFactory factory = TableFactory.getInstance();
    table=factory.getSymbolTable();
    info = new Hashtable();
    PrintWriter pw=new PrintWriter(new BufferedWriter(new FileWriter("prr")));
    c=new Code(pw);
    l=new Label();
  }


  //-------------------------------------------------

  public void inAProgram(AProgram node){
     table.enterBlock();
     try{
       type_int = (Type)table.lookup("int");
       type_int.setVal(1);

       type_real = (Type)table.lookup("real");
       type_real.setVal(1);

       type_char = (Type)table.lookup("char");
       type_char.setVal(1);

       type_bool = (Type)table.lookup("bool");
       type_bool.setVal(1);

       type_string = (Type)table.lookup("string");
     }catch(TableError e){
       System.out.println(e.getMessage());
    }
  }

  public void outAProgram(AProgram node){
   try{
      String ident = ((Information2)info.get(node.getProgramHead())).getString();
      double label = (double)((Program)table.lookup(ident)).getLabel();
      Codeptr block_code = ((Information2)info.get(node.getBlock())).getCode();
      Codeptr code = Code.concat(block_code, Code.gen(Opecode.o_q), 
                                 Code.gen(Opecode.o_mst), 
                                 Code.gen(Opecode.o_cup, null_val, label, Null_Val),
                                 Code.gen(Opecode.o_stp));
      Code.printcode(code);
      c.linearize(code);
      table.exitBlock();
    }catch(Exception e){
      System.out.println(e.getMessage());
    }
  }

  public void outAProgramHead(AProgramHead node){
    try{
      Information2 ident = (Information2)info.get(node.getIdent());
      hasParameter proc = (hasParameter)table.lookup(ident.getString());
    }catch(TableError e){
      System.out.println(e.getMessage());
    }
    info.put(node, info.get(node.getIdent()));
  }

  public void inABlock(ABlock node){
    if(! ( node.parent() instanceof AProcDecl || node.parent() instanceof AFuncDecl)){
      table.enterBlock();
      Information2 block = new Information2();
      block.set(Code.nullcode());
      info.put(node, block);
    }else{
      info.put(node, info.get(node.parent()));
    }
  }

  public void outABlock(ABlock node){
    table.exitBlock();
    Information2 block = (Information2)info.get(node);
    Codeptr decl_code = ((Information2)info.get(node.getConstTypeVarProcFuncDeclPart())).getCode();
    Codeptr stmt_code = ((Information2)info.get(node.getStmtPart())).getCode();
    Codeptr code = Code.concat(decl_code, stmt_code);
    block.set(code);
    info.put(node, block);
  }

  public void outAConstTypeVarProcFuncDeclPart(AConstTypeVarProcFuncDeclPart node){
    hasParameter proc = (hasParameter)table.getCurrent();
    if(proc instanceof Program){
      proc.setLabel(l.getLabel());
    }
    info.put(node, info.get(node.getProcFuncDeclPart()));
  }

  public void outAVarDecl(AVarDecl node){
    try{
      Information2 ident_list = (Information2)info.get(node.getIdentList());
      Iterator it = ident_list.getList().iterator();
      while(it.hasNext()){
	Variable var = (Variable)table.lookup(((Information2)it.next()).getString());
	int val = var.getType().getVal();
	int address = addVal1(val, (hasParameter)table.getCurrent());
	var.setAddress(address);
      }
    }catch(Exception e){
      System.out.println(e.getMessage());
    }
  }

  public void outAIdentListIdentList(AIdentListIdentList node){
    Information2 ident_list = (Information2)info.get(node.getIdentList());
    ConcreteAggregate list = ident_list.getList();
    list.add(info.get(node.getIdent()));
    info.put(node, ident_list);
  }

  public void outAIdentIdentList(AIdentIdentList node){
    Information2 ident_list = new Information2();
    ConcreteAggregate list = new ConcreteAggregate();
    list.add(info.get(node.getIdent()));
    ident_list.set(list);
    info.put(node, ident_list);
  }

  public void outAIdent(AIdent node){
    Information2 ident = new Information2();
    ident.set(node.getIdentt().getText().trim());
    info.put(node, ident);
  }

  public void outAProcFuncDeclListProcFuncDeclPart
                  (AProcFuncDeclListProcFuncDeclPart node){
    info.put(node, info.get(node.getProcFuncDeclList()));
  }

  public void outAEmptyProcFuncDeclPart(AEmptyProcFuncDeclPart node){
    Information2 func_decl = new Information2();
    func_decl.set(Code.nullcode());
    info.put(node, func_decl); 
 }

  public void outAProcFuncDeclListProcFuncDeclList
                  (AProcFuncDeclListProcFuncDeclList node){
    Information2 proc_func_decl_list = new Information2();
    Information2 decl_list= (Information2)info.get(node.getProcFuncDeclList());
    Information2 decl = (Information2)info.get(node.getProcFuncDecl());
    Codeptr code = Code.concat(decl_list.getCode(), decl.getCode());
    proc_func_decl_list.set(code);
    info.put(node, proc_func_decl_list);
  }

  public void outAProcFuncDeclProcFuncDeclList
                  (AProcFuncDeclProcFuncDeclList node){
    info.put(node, info.get(node.getProcFuncDecl()));
  }

  public void outAProcDeclProcFuncDecl(AProcDeclProcFuncDecl node){
    info.put(node, info.get(node.getProcDecl()));
  }

  public void outAFuncDeclProcFuncDecl(AFuncDeclProcFuncDecl node){
    info.put(node, info.get(node.getFuncDecl()));
  }

  public void outAProcDecl(AProcDecl node){
    info.put(node, info.get(node.getBlock()));
  }

  public void outAProcDeclHead(AProcDeclHead node){
    try{
      Information2 decl_head = new Information2();
      String ident=((Information2)info.get(node.getIdent())).getString();
      int label = l.getLabel();
      Procedure proc = (Procedure)table.lookup(ident);
      proc.setLabel(label);

      Codeptr code = Code.nullcode();
      table.enterBlock();
      Iterator it = proc.getParameter();
      while(it.hasNext()){
        Parameter fp = (Parameter)it.next();
        int address = addVal1(1, ((hasParameter)proc));
        fp.setAddress(address);

        if(fp instanceof ValueParameter
            &&(fp.getType() instanceof ArrayType 
              || fp.getType() instanceof RecordType)){
	  address = addVal1(fp.getType().getVal(), proc);
          code = Code.concat(Code.gen(Opecode.o_lda, null_val, 0.0, address),
	  	              Code.gen(Opecode.o_lod, Codetype.c_addr, 0.0, fp.getAddress()),
			      Code.gen(Opecode.o_mov, null_val, fp.getType().getVal(), Null_Val),
			      code);
	    fp.setAddress(address);
        }
      }
      decl_head.set(code);
      info.put(node.parent(), decl_head);
    }catch(Exception e){
      System.out.println(e.getMessage());
    }
  }

  public void outAFuncDecl(AFuncDecl node){
    info.put(node, info.get(node.getBlock()));
  }

  public void outAFuncDeclHead(AFuncDeclHead node){
    try{
      Information2 decl_head = new Information2();
      String ident=((Information2)info.get(node.getIdent())).getString();
      int label = l.getLabel();
      Function func = (Function)table.lookup(ident);
      func.setLabel(label);

      Codeptr code = Code.nullcode();
      table.enterBlock();
      Iterator it = func.getParameter();
      while(it.hasNext()){
        Parameter fp = (Parameter)it.next();
        int address = addVal1(1, func);
        fp.setAddress(address);

        if(fp instanceof ValueParameter
                 &&(fp.getType() instanceof ArrayType 
                   || fp.getType() instanceof RecordType)){
        address = (int)addVal1(fp.getType().getVal(), func);
        code = Code.concat(Code.gen(Opecode.o_lda, null_val, 0.0, address),
			    Code.gen(Opecode.o_lod, Codetype.c_addr, 0.0, (double)fp.getAddress()),
	                    Code.gen(Opecode.o_mov, null_val, fp.getType().getVal(), Null_Val),
			    code);
	    fp.setAddress(address);
        }
      }
      decl_head.set(code);
      info.put(node.parent(), decl_head);
    }catch(Exception e){
      System.out.println(e.getMessage());
    }
 }

  public void inAStmtPart(AStmtPart node){
    Information2 stmt = new Information2();
    stmt.set(l.getLabel(), l.getLabel());
    info.put(node, stmt);
  }

  public void outAStmtPart(AStmtPart node){
    Information2 stmt_seq = (Information2)info.get(node.getStmtSeq());
    Information2 param_list = (Information2)info.get(node.parent());
    Information2 stmt_part = (Information2)info.get(node);
    double plab=(double)stmt_part.getLab();
    double plab2=(double)stmt_part.getLab2();
    hasParameter proc = (hasParameter)table.getCurrent();
    double label = (double)proc.getLabel();
    double retval=0.0;
    if(proc instanceof Program){
      retval = 9.0 + proc.getVal2();
    }else{
      retval = 5.0+proc.getVal2();
    }
    Codeptr code=Code.concat(Code.gen(Opecode.o_lab, 0, label, Null_Val), 
			     Code.gen(Opecode.o_ent,1, plab, Null_Val),
			     Code.gen(Opecode.o_ent, 2, plab2, Null_Val),
                             param_list.getCode(),
			     stmt_seq.getCode(),
                             retCode(proc),
                             Code.gen(Opecode.o_lab, 1, plab, retval), 
                             Code.gen(Opecode.o_lab, 1, plab2, Null_Val));
    stmt_part.set(code);
  }


  public void outAArrayType(AArrayType node){
    try{
      String name 
        = new Integer(node.getArray().getLine()).toString()
               + "-" + new Integer(node.getArray().getPos()).toString();
      ArrayType array_type = (ArrayType)table.lookup(name);
      int address = (int)array_type.getType().getVal();
      ConcreteAggregate index_list = array_type.getIndex();
      for(int i = index_list.getLength()-1 ; i >= 0; i--){
        Index index = (Index)index_list.get(i);
        int range = index.getRange();
        index.setAddress(address);
        address = range * address;
      }
      array_type.setVal(address);
    }catch(TableError e){
      System.out.println(e.getMessage());
    }
  }

  public void outARecordType(ARecordType node){
    try{
      String name 
        = new Integer(node.getRecord().getLine()).toString() 
                + "-" + new Integer(node.getRecord().getPos()).toString();
      RecordType record_type = (RecordType)table.lookup(name);
      table.enterBlock();

      int address = 0;
      ConcreteAggregate list = ((Information2)info.get(node.getRecordBodyList())).getList();
      Iterator it = list.iterator();
      while(it.hasNext()){
        ConcreteAggregate ident_list = ((Information2)it.next()).getList();
        Iterator ident_it = ident_list.iterator();
        while(ident_it.hasNext()){
          Field field = (Field)table.lookup(((Information2)ident_it.next()).getString(), 
                                            record_type.getField());
          int field_type_val = field.getType().getVal();
          field.setAddress(address);
          address = address + field_type_val;
        }
      }
      table.exitBlock();

      record_type.setVal(address);
    }catch(TableError e){
      System.out.println(e.getMessage());
    }
 }

  public void outARecordBodyListRecordBodyList(ARecordBodyListRecordBodyList node){
    Information2 body_list = (Information2)info.get(node.getRecordBodyList());
    ConcreteAggregate list = body_list.getList();
    list.add(info.get(node.getRecordBody()));
    info.put(node, body_list);
  }

  public void outARecordBodyRecordBodyList(ARecordBodyRecordBodyList node){
    Information2 rec_body = new Information2();
    ConcreteAggregate list = new ConcreteAggregate();
    list.add(info.get(node.getRecordBody()));
    rec_body.set(list);
    info.put(node, rec_body);
  }

  public void outARecordBody(ARecordBody node){
    info.put(node, info.get(node.getIdentList()));
  }

  public void outAStmtSeqStmtSeq(AStmtSeqStmtSeq node){
    Information2 s_seq = new Information2();
    Codeptr stmt_seq = ((Information2)info.get(node.getStmtSeq())).getCode();
    Codeptr stmt = ((Information2)info.get(node.getStmt())).getCode();
    Codeptr code = Code.concat(stmt_seq, stmt);
    s_seq.set(code);
    info.put(node, s_seq);
  }

  public void outAStmtStmtSeq(AStmtStmtSeq node){
    info.put(node, info.get(node.getStmt()));
  }

  public void outAAssignmentStmt(AAssignmentStmt node){
    info.put(node, info.get(node.getAssignment()));
  }

  public void outAProcStmt(AProcStmt node){
    info.put(node, info.get(node.getProcCall()));
  }

  public void outAStmtSeqStmt(AStmtSeqStmt node){
      info.put(node, info.get(node.getStmtSeq()));
  }

  public void outAIfStmtStmt(AIfStmtStmt node){
    info.put(node, info.get(node.getIfStmt()));
  }

  public void outACaseStmtStmt(ACaseStmtStmt node){   
    info.put(node, info.get(node.getCaseStmt()));
  }

  public void outAWhileStmtStmt(AWhileStmtStmt node){
    info.put(node, info.get(node.getWhileStmt()));
  }

  public void outARepeatStmtStmt(ARepeatStmtStmt node){
    info.put(node, info.get(node.getRepeatStmt()));
  }

  public void outAForStmtStmt(AForStmtStmt node){   
    info.put(node, info.get(node.getForStmt()));
  }

  public void outAEmptyStmt(AEmptyStmt node){
    Information2 empty_stmt = new Information2();
    empty_stmt.set(Code.nullcode());
    info.put(node, empty_stmt);
  }

  public void outAVariableTailAssignment(AVariableTailAssignment node){
    Information2 assignment = new Information2();
    String ident = ((Information2)info.get(node.getIdent())).getString();
    Information2 expr=(Information2)info.get(node.getExpr());
    Codeptr expr_code=expr.getCode();
    Type expr_type=expr.getType();
    ConcreteAggregate variable_tail = ((Information2)info.get(node.getVariableTail())).getList();
    Codeptr code=vtailCodeL(ident, variable_tail, expr_code, expr_type, table.getLevel());
    assignment.set(code);
    info.put(node, assignment);
  }

  public void outAColeqAssignment(AColeqAssignment node){
    try{
      Information2 assignment = new Information2();
      String ident = ((Information2)info.get(node.getIdent())).getString();
      Entry entry = table.lookup(ident);
      Type type = ((hasType)entry).getType();

      Information2 expr=(Information2)info.get(node.getExpr());
      Codeptr expr_code = expr.getCode();
      Type expr_type = expr.getType();

      Codeptr code2_3 = null;
      if(type == type_bool){
        code2_3 = Code.gen(Opecode.o_chk, Codetype.c_boolean, 0.0, 1.0);
      }else if(type == type_char){
        code2_3 = Code.gen(Opecode.o_chk, Codetype.c_character, 0.0, 63.0);
      }else if(type == type_real &&  expr_type == type_int){
        code2_3 = Code.gen(Opecode.o_flt);
      }else{
        code2_3 = Code.nullcode();
      }

      Codeptr assignment_code=null;	  
      if(entry instanceof VariableParameter){
	assignment_code = 
              Code.concat(Code.gen(Opecode.o_lod, Codetype.c_addr, 
                          Null_Val, ((Parameter)entry).getAddress()),
                          expr_code, 
                          code2_3, 
		          Code.gen(Opecode.o_sto, retType(type), Null_Val, Null_Val));
      }else if(entry instanceof Function){
	assignment_code = Code.concat(expr_code, 
                                       code2_3, 
				       Code.gen(Opecode.o_str, retType(type), Null_Val, Null_Val));
      }else{
	if(entry.getLevel() == 1){
	  assignment_code = Code.concat(expr_code, 
                                         code2_3,
					 Code.gen(Opecode.o_sro, retType(type), 
						  ((hasAddress)entry).getAddress(), Null_Val));
	}else{
	  assignment_code=Code.concat(expr_code, 
                                       code2_3,
				       Code.gen(Opecode.o_str, retType(type), 
                                              table.getLevel() - entry.getLevel(), 
                                              ((hasAddress)entry).getAddress()));
	}
      }
      assignment.set(assignment_code);
      info.put(node, assignment);
    }catch(TableError e){
      System.out.println(e.getMessage());
    }
  }

  public void outABuiltinProcCallProcCall(ABuiltinProcCallProcCall node){
    info.put(node, info.get(node.getBuiltinProcCall()));
  }

  public void outAActualParamListProcCall(AActualParamListProcCall node){
    String ident = ((Information2)info.get(node.getIdent())).getString();
    ConcreteAggregate param_list = ((Information2)info.get(node.getActualParamList())).getList();
    Information2 proc_call = funcwParamCode(ident, param_list, table.getLevel());
    info.put(node, proc_call);
  }

  public void outAIdentProcCall(AIdentProcCall node){
    String ident = ((Information2)info.get(node.getIdent())).getString();
    Information2 proc_call = funcwParamCode(ident, new ConcreteAggregate(), table.getLevel());
    info.put(node, proc_call);
  }

  public void outAReadBuiltinProcCall(AReadBuiltinProcCall node){
    Information2 proc_call = new Information2();
    ConcreteAggregate list = ((Information2)info.get(node.getActualParamList())).getList();
    Iterator it = list.iterator();
    Codeptr code = Code.nullcode();
    while(it.hasNext()){
      Information2 expr = (Information2)it.next();
      code = Code.concat(code, changeCode(expr.getCode()), 
                          Code.gen(Opecode.o_lda, null_val, table.getLevel()-1, 5.0),
                          Code.gen(Opecode.o_csp, Builtinfunc.read, retType(expr.getType()), Null_Val));
    }
    proc_call.set(code);
    info.put(node, proc_call);
  }

  public void outAReadlnBuiltinProcCall(AReadlnBuiltinProcCall node){
    Information2 proc_call = new Information2();
    ConcreteAggregate list = ((Information2)info.get(node.getActualParamList())).getList();
    Iterator it = list.iterator();
    Codeptr code = Code.nullcode();
    while(it.hasNext()){
      Information2 expr = (Information2)it.next();
      code = Code.concat(code, changeCode(expr.getCode()), 
                          Code.gen(Opecode.o_lda, null_val, table.getLevel()-1, 5.0),
                          Code.gen(Opecode.o_csp, Builtinfunc.read, retType(expr.getType()), Null_Val));
    }
    code = Code.concat(code, 
                        Code.gen(Opecode.o_lda, null_val, table.getLevel() - 1, Null_Val),
                        Code.gen(Opecode.o_csp, Builtinfunc.readln, Null_Val, Null_Val));
    proc_call.set(code);
    info.put(node, proc_call);
  }

  public void outAWriteBuiltinProcCall(AWriteBuiltinProcCall node){
    Information2 proc_call = new Information2();
    ConcreteAggregate list = ((Information2)info.get(node.getActualParamList())).getList();
    Iterator it = list.iterator();
    Codeptr code = Code.nullcode();
    while(it.hasNext()){
      Information2 expr = (Information2)it.next();
      Type expr_type = expr.getType();
      Codeptr expr_code = expr.getCode();
      if(expr_type == type_string){
        code = Code.concat(code, expr_code, 
                            Code.gen(Opecode.o_lda, null_val, table.getLevel()-1, 6.0),
                            Code.gen(Opecode.o_csp, Builtinfunc.write, Codetype.c_string, Null_Val));
      }else{
        double t=0.0;
        if(expr_type==type_char) t=1.0;
        else if(expr_type==type_real) t=20.0;
        else t=10.0;
 
        code = Code.concat(code, cutCode(expr_code),
                           Code.gen(Opecode.o_ldc, Codetype.c_integer,t, Null_Val),
                           Code.gen(Opecode.o_lda, null_val, table.getLevel()-1, 6.0),
	                   Code.gen(Opecode.o_csp, Builtinfunc.write, retType(expr_type), Null_Val));
      }
    }
    proc_call.set(code);
    info.put(node, proc_call);
  }

  public void outAWritelnBuiltinProcCall(AWritelnBuiltinProcCall node){
    Information2 proc_call = new Information2();
    Codeptr code = Code.concat(Code.gen(Opecode.o_lda, null_val, table.getLevel() - 1, 6.0),
                               Code.gen(Opecode.o_csp, Builtinfunc.writeln, Null_Val, Null_Val));
    proc_call.set(code);
    info.put(node, proc_call);
  }

  public void outAWritelnlparBuiltinProcCall(AWritelnlparBuiltinProcCall node){
    Information2 proc_call = new Information2();
    ConcreteAggregate list = ((Information2)info.get(node.getActualParamList())).getList();
    Iterator it = list.iterator();
    Codeptr code = Code.nullcode();
    while(it.hasNext()){
      Information2 expr = (Information2)it.next();
      Type expr_type = expr.getType();
      Codeptr expr_code = expr.getCode();
      if(expr_type == type_string){
        code = Code.concat(code, expr_code, 
                            Code.gen(Opecode.o_lda, null_val, table.getLevel()-1, 6.0),
                            Code.gen(Opecode.o_csp, Builtinfunc.write, Codetype.c_string, Null_Val));
      }else{
        double t=0.0;
        if(expr_type==type_char) t=1.0;
        else if(expr_type==type_real) t=20.0;
        else t=10.0;
 
        code = Code.concat(code, cutCode(expr_code),
                           Code.gen(Opecode.o_ldc, Codetype.c_integer,t, Null_Val),
                           Code.gen(Opecode.o_lda, null_val, table.getLevel()-1, 6.0),
	                   Code.gen(Opecode.o_csp, Builtinfunc.write, retType(expr_type), Null_Val));
      }
    }
    code = Code.concat(code, 
                        Code.gen(Opecode.o_lda, null_val, table.getLevel() - 1, 6.0),
                        Code.gen(Opecode.o_csp, Builtinfunc.writeln, Null_Val, Null_Val));
    proc_call.set(code);
    info.put(node, proc_call);
  }

  public void outAMatchedStmtIfStmt(AMatchedStmtIfStmt node){
    info.put(node, info.get(node.getFmatchedStmt()));
  }

  public void outAUnmatchedStmtIfStmt(AUnmatchedStmtIfStmt node){
    info.put(node, info.get(node.getUnmatchedStmt()));
  }

  public void inAFmatchedStmt(AFmatchedStmt node){
    Information2 if_stmt = new Information2();
    if_stmt.set(l.getLabel(), l.getLabel());
    info.put(node, if_stmt);
  }

  public void outAFmatchedStmt(AFmatchedStmt node){
    Codeptr right_code = ((Information2)info.get(node.getRight())).getCode();
    Codeptr left_code = ((Information2)info.get(node.getLeft())).getCode();
    Codeptr expr_code = ((Information2)info.get(node.getExpr())).getCode();
    Information2 if_stmt = (Information2)info.get(node);
    int ilab = if_stmt.getLab();
    int ilab2 = if_stmt.getLab2();
    Codeptr code = Code.concat(expr_code, 
                                Code.gen(Opecode.o_fjp, ilab, Null_Val, Null_Val),
                                left_code, 
                                Code.gen(Opecode.o_ujp, ilab2, Null_Val ,Null_Val),
                                Code.gen(Opecode.o_lab, null_val, ilab, Null_Val),
                                right_code, 
                                Code.gen(Opecode.o_lab, null_val, ilab2, Null_Val));
    if_stmt.set(code);
  }

  public void inAIfthenIfUnmatchedStmt(AIfthenIfUnmatchedStmt node){
    Information2 if_stmt = new Information2();
    if_stmt.set(l.getLabel());
    info.put(node, if_stmt);
  }

  public void outAIfthenIfUnmatchedStmt(AIfthenIfUnmatchedStmt node){
    Codeptr stmt_code = ((Information2)info.get(node.getIfStmt())).getCode();
    Codeptr expr_code = ((Information2)info.get(node.getExpr())).getCode();
    Information2 if_stmt = (Information2)info.get(node);
    int ilab = if_stmt.getLab();
    Codeptr code = Code.concat(expr_code, 
                                Code.gen(Opecode.o_fjp, ilab, Null_Val, Null_Val),
                                stmt_code, 
                                Code.gen(Opecode.o_lab, 0, ilab,Null_Val));
    if_stmt.set(code);
  }

  public void inAIfthenUnmatchedStmt(AIfthenUnmatchedStmt node){
    Information2 if_stmt = new Information2();
    if_stmt.set(l.getLabel());
    info.put(node, if_stmt);
  }

  public void outAIfthenUnmatchedStmt(AIfthenUnmatchedStmt node){
    Codeptr stmt_code = ((Information2)info.get(node.getOther())).getCode();
    Codeptr expr_code = ((Information2)info.get(node.getExpr())).getCode();
    Information2 if_stmt = (Information2)info.get(node);
    int ilab = if_stmt.getLab();
    Codeptr code = Code.concat(expr_code, 
                                Code.gen(Opecode.o_fjp, ilab, Null_Val, Null_Val),
                                stmt_code, 
                                Code.gen(Opecode.o_lab, 0, ilab, Null_Val));
    if_stmt.set(code);
  }

  public void inAIfthenelseUnmatchedStmt(AIfthenelseUnmatchedStmt node){
    Information2 if_stmt = new Information2();
    if_stmt.set(l.getLabel(), l.getLabel());
    info.put(node, if_stmt);
  }

  public void outAIfthenelseUnmatchedStmt(AIfthenelseUnmatchedStmt node){
    Codeptr right_code = ((Information2)info.get(node.getUnmatchedStmt())).getCode();
    Codeptr left_code = ((Information2)info.get(node.getSmatchedStmt())).getCode();
    Codeptr expr_code = ((Information2)info.get(node.getExpr())).getCode();
    Information2 if_stmt = (Information2)info.get(node);
    int ilab = if_stmt.getLab();
    int ilab2 = if_stmt.getLab2();
    Codeptr code = Code.concat(expr_code, 
                                Code.gen(Opecode.o_fjp, ilab, Null_Val, Null_Val),
                                left_code, 
                                Code.gen(Opecode.o_ujp, ilab2, Null_Val ,Null_Val),
                                Code.gen(Opecode.o_lab, null_val, ilab, Null_Val),
                                right_code, 
                                Code.gen(Opecode.o_lab, null_val, ilab2, Null_Val));
    if_stmt.set(code);
  }

  public void outAFmatchedStmtSmatchedStmt(AFmatchedStmtSmatchedStmt node){
    info.put(node, info.get(node.getFmatchedStmt()));
  }

  public void outAOtherSmatchedStmt(AOtherSmatchedStmt node){
    info.put(node, info.get(node.getOther()));
  }

  public void outAAssignmentOther(AAssignmentOther node){
    info.put(node, info.get(node.getAssignment()));
  }

  public void outAProcOther(AProcOther node){
    info.put(node, info.get(node.getProcCall()));
  }

  public void outAStmtSeqOther(AStmtSeqOther node){
    info.put(node, info.get(node.getStmtSeq()));
  }

  public void outACaseStmtOther(ACaseStmtOther node){
    info.put(node, info.get(node.getCaseStmt()));
  }

  public void outARepeatStmtOther(ARepeatStmtOther node){
    info.put(node, info.get(node.getRepeatStmt()));
  }

  public void outAEmptyOther(AEmptyOther node){
    info.put(node, c.nullcode());
  }

  public void inACaseStmt(ACaseStmt node){
    Information2 case_stmt = new Information2();
    case_stmt.set(l.getLabel(), l.getLabel());
    info.put(node, case_stmt);
  }

  public void outACaseStmt(ACaseStmt node){
    TableFactory factory = TableFactory.getInstance();
    CaseLabelTable case_label_table  
        = factory.getCaseLabelTable(node.getCase().getLine(), node.getCase().getPos());

    Information2 case_stmt = (Information2)info.get(node);
    int clab=case_stmt.getLab();
    int clab2=case_stmt.getLab2();
    int clab3=l.getLabel();

    Information2 expr=(Information2)info.get(node.getExpr());
    Codeptr code1 = expr.getCode();
    Type expr_type = expr.getType();

    Codeptr code2 = null;
    if(expr_type != type_int){
      code2 = Code.gen(Opecode.o_ord, retType(expr_type), Null_Val,Null_Val);
    }else{
      code2 = Code.nullcode();
    }


    Information2 case_body_list = (Information2)info.get(node.getCaseBodyList());
    ConcreteAggregate list = case_body_list.getList();
    Iterator it = list.iterator();
    Codeptr code4 = Code.nullcode();
    while(it.hasNext()){
      Information2 case_body = (Information2)it.next();
      Codeptr stmt_code = case_body.getCode();
      int body_label = case_body.getLab();

      ConcreteAggregate case_label_list = case_body.getList();
      if(case_label_list == null){
        break;
      }
      Iterator it2 = case_label_list.iterator();
      while(it2.hasNext()){
	try{
          Information2 case_label = (Information2)it2.next();
          CaseLabel entry = (CaseLabel)case_label_table.lookup(case_label.getString());
          entry.setLabel(body_label);
	}catch(TableError e){
	  System.out.println(e.getMessage());
	}
      }
      code4 = Code.concat(code4,
                           Code.concat(Code.gen(Opecode.o_lab, 0, body_label, Null_Val), 
                                        stmt_code,
	                                Code.gen(Opecode.o_ujp, clab2, Null_Val, Null_Val)));
    }

    Codeptr code6_code6 = Code.nullcode();
    Iterator it3 = case_label_table.iterator();
    int i = case_label_table.getMin();
    while(it3.hasNext()){
      CaseLabel caselabel = (CaseLabel)it3.next();
      int label_order = caselabel.getOrder();
      while(i != label_order){
        code6_code6 = Code.concat(code6_code6,Code.gen(Opecode.o_ujc, null_val, Null_Val,Null_Val));
        i++;
      }
      code6_code6 = Code.concat(code6_code6,
		                 Code.gen(Opecode.o_ujp, caselabel.getLabel(), Null_Val, Null_Val));
      i++;
    }

    Codeptr code6 = Code.concat(Code.gen(Opecode.o_chk, Codetype.c_integer, 
                                      (double)case_label_table.getMin(), case_label_table.getMax()),
			         Code.gen(Opecode.o_ldc, Codetype.c_integer,
                                      case_label_table.getMin(), Null_Val),
			         Code.gen(Opecode.o_sb, Codetype.c_integer, Null_Val, Null_Val),
			         Code.gen(Opecode.o_xjp, clab3, Null_Val, Null_Val),
			         Code.gen(Opecode.o_lab, 0, clab3, Null_Val),
			         code6_code6);      


    Codeptr code=Code.concat(code1, code2,
	      	              Code.gen(Opecode.o_ujp, clab, Null_Val, Null_Val),
                              code4,
		              Code.gen(Opecode.o_lab, 0, clab,Null_Val),
		              code6,
		              Code.gen(Opecode.o_lab, 0, clab2, Null_Val));
    case_stmt.set(code);
  }

  public void outACaseBodyListCaseBodyList(ACaseBodyListCaseBodyList node){
    Information2 case_body_list = (Information2)info.get(node.getCaseBodyList());
    ConcreteAggregate list = case_body_list.getList();
    list.add(info.get(node.getCaseBody()));
    info.put(node, case_body_list);
  }

  public void outACaseBodyCaseBodyList(ACaseBodyCaseBodyList node){
    Information2 case_list = new Information2();
    ConcreteAggregate case_body_list = new ConcreteAggregate();
    case_body_list.add(info.get(node.getCaseBody()));
    case_list.set(case_body_list);
    info.put(node, case_list);
  }

  public void inACaseLabelListCaseBody(ACaseLabelListCaseBody node){
    Information2 case_label = new Information2();
    case_label.set(l.getLabel());
    info.put(node,  case_label);
  }

  public void outACaseLabelListCaseBody(ACaseLabelListCaseBody node){
    Information2 case_body = (Information2)info.get(node);
    
    Codeptr stmt_code = ((Information2)info.get(node.getStmt())).getCode();
    case_body.set(stmt_code);

    ConcreteAggregate case_label_list = ((Information2)info.get(node.getCaseLabelList())).getList();
    case_body.set(case_label_list);
  }  


  public void outAEmptyCaseBody(AEmptyCaseBody node){
    Information2 empty = new Information2();
    empty.set(Code.nullcode());
    info.put(node, empty);
  }

  public void outACaseLabeListCaseLabelList(ACaseLabelListCaseLabelList node){
    Information2 case_label_list = (Information2)info.get(node.getCaseLabelList());
    ConcreteAggregate list = case_label_list.getList();
    list.add(info.get(node.getCaseLabel()));
    info.put(node, case_label_list);
  }

  public void outACaseLabelCaseLabelList(ACaseLabelCaseLabelList node){
    Information2 case_label = new Information2();
    ConcreteAggregate case_label_list = new ConcreteAggregate();
    case_label_list.add(info.get(node.getCaseLabel()));
    case_label.set(case_label_list);
    info.put(node, case_label);
  }

  public void outACaseLabel(ACaseLabel node){
    info.put(node, info.get(node.getConstant()));
  }

  public void inAWhileStmt(AWhileStmt node){
    Information2 while_stmt = new Information2();
    while_stmt.set(l.getLabel(),l.getLabel());
    info.put(node, while_stmt);
  }

  public void outAWhileStmt(AWhileStmt node){
    Information2 stmt = (Information2)info.get(node.getStmt());
    Codeptr stmt_code = stmt.getCode();
    Information2 expr = (Information2)info.get(node.getExpr());
    Codeptr expr_code = expr.getCode();
    Information2 while_stmt = (Information2)info.get(node);
    int wlab = while_stmt.getLab();
    int wlab2 = while_stmt.getLab2();
    Codeptr code=Code.concat(Code.gen(Opecode.o_lab, null_val, wlab, Null_Val),
                              expr_code, 
                              Code.gen(Opecode.o_fjp, wlab2, Null_Val, Null_Val),
                              stmt_code, 
                              Code.gen(Opecode.o_ujp,  wlab, Null_Val, Null_Val),
                              Code.gen(Opecode.o_lab, null_val, wlab2, Null_Val));
    while_stmt.set(code);
  }

  public void inARepeatStmt(ARepeatStmt node){
    Information2 repeat_stmt = new Information2();
    repeat_stmt.set(l.getLabel());
    info.put(node, repeat_stmt);
  }

  public void outARepeatStmt(ARepeatStmt node){
    Information2 stmt = (Information2)info.get(node.getStmtSeq());
    Codeptr stmt_code = stmt.getCode();
    Information2 expr = (Information2)info.get(node.getExpr());
    Codeptr expr_code = expr.getCode();
    Information2 repeat_stmt = (Information2)info.get(node);
    int rlab = repeat_stmt.getLab();
    Codeptr code = Code.concat(Code.gen(Opecode.o_lab, null_val, rlab, Null_Val),
                                stmt_code, expr_code,
                                Code.gen(Opecode.o_fjp,  rlab, Null_Val, Null_Val));
    repeat_stmt.set(code);
  }

  public void inAForStmt(AForStmt node){
    Information2 for_stmt = new Information2();
    for_stmt.set(l.getLabel(), l.getLabel());
    info.put(node, for_stmt);
  }

  public void outAForStmt(AForStmt node){
    try{
      Information2 left = (Information2)info.get(node.getLeft());
      Codeptr left_code = left.getCode();

      Information2 right = (Information2)info.get(node.getRight());
      Codeptr right_code = right.getCode();

      Information2 stmt = (Information2)info.get(node.getStmt());
      Codeptr stmt_code = stmt.getCode();

      Information2 ident = (Information2)info.get(node.getIdent());
      Entry entry = table.lookup(ident.getString());

      Information2 for_stmt = (Information2)info.get(node);
      int plab=for_stmt.getLab();
      int plab2=for_stmt.getLab2();

      Codeptr code1 = null;
      Codeptr code2 = null;
      if(entry.getLevel() == 1){
	code1 = Code.gen(Opecode.o_sro, Codetype.c_integer, ((hasAddress)entry).getAddress(), Null_Val);
	code2 = Code.gen(Opecode.o_ldo, Codetype.c_integer, ((hasAddress)entry).getAddress(), Null_Val);
      }else{
	code1 = Code.gen(Opecode.o_str, Codetype.c_integer,
		   table.getLevel() - entry.getLevel(), ((hasAddress)entry).getAddress());
	code2 = Code.gen(Opecode.o_lod, Codetype.c_integer,
		   table.getLevel() - entry.getLevel(), ((hasAddress)entry).getAddress());
      }

      hasParameter proc = (hasParameter)table.getCurrent();
      int add = 0;
      if(proc instanceof Program){
        add = proc.getVal1() + 9;
      }else{
        add = proc.getVal1() + 5;
      }

      Codeptr op1=null, op2=null;
      if(node.getDirection() instanceof AToDirection){
       op1=Code.gen(Opecode.o_leq, Codetype.c_integer, Null_Val, Null_Val);
       op2=Code.gen(Opecode.o_inc, Codetype.c_integer, 1.0, Null_Val);
      }else{
       op1=Code.gen(Opecode.o_geq, Codetype.c_integer, Null_Val, Null_Val);
       op2=Code.gen(Opecode.o_dec,Codetype.c_integer, 1.0, Null_Val);
      }
 
      Codeptr code=Code.concat(left_code, 
                                Code.concat(code1, right_code, 
                                             Code.gen(Opecode.o_str, Codetype.c_integer, 0.0, add),
                                             Code.gen(Opecode.o_lab, 0, plab, Null_Val), code2, 
                                             Code.gen(Opecode.o_lod, Codetype.c_integer, 0.0, add), 
                                             op1),
                               Code.concat(Code.gen(Opecode.o_fjp, plab2, Null_Val, Null_Val), 
                                            stmt_code,  code2.copy(), op2, code1.copy(),
                                            Code.gen(Opecode.o_ujp, plab, Null_Val, Null_Val), 
                                            Code.gen(Opecode.o_lab, 0, plab2, Null_Val)));
      for_stmt.set(code);
    }catch(TableError e){
      System.out.println(e.getMessage());
    }
  }

  public void outASimpleExprExpr(ASimpleExprExpr node){
    info.put(node, info.get(node.getSimpleExpr()));
  }

  public void outASimpleExprrelopsimpleExprExpr(ASimpleExprrelopsimpleExprExpr node){
    Information2 expr = new Information2();
    Information2 left = (Information2)info.get(node.getLeft());
    Codeptr left_code = left.getCode();
    Type left_type = left.getType();

    Information2 right = (Information2)info.get(node.getRight());
    Codeptr right_code = right.getCode();
    Type right_type = right.getType();

    Type type = null;
    Codeptr code = null;
    if(left_type == type_int && right_type == type_int){
      type=type_int;
      code = Code.nullcode();
    }else{
      type=type_real;
      if(left_type == type_int){
        code = Code.gen(Opecode.o_flo);
      }else if(right_type == type_int){
        code = Code.gen(Opecode.o_flt);
      }else{
        code = Code.nullcode();
      }
    }

    if(node.getRelop() instanceof AEqRelop){
      code = Code.concat(code, Code.gen(Opecode.o_equ, retType(type), Null_Val, Null_Val));
    }else if(node.getRelop() instanceof ANoteqRelop){
      code = Code.concat(code, Code.gen(Opecode.o_neq, retType(type), Null_Val, Null_Val));  
    }else if(node.getRelop() instanceof AGtRelop){
      code = Code.concat(code, Code.gen(Opecode.o_grt, retType(type), Null_Val, Null_Val));
    }else if(node.getRelop() instanceof ALtRelop){
      code = Code.concat(code, Code.gen(Opecode.o_les, retType(type), Null_Val, Null_Val));
    }else if(node.getRelop() instanceof AGeRelop){
      code = Code.concat(code, Code.gen(Opecode.o_geq, retType(type), Null_Val, Null_Val));
    }else if(node.getRelop() instanceof ALeRelop){
      code = Code.concat(code, Code.gen(Opecode.o_leq, retType(type), Null_Val, Null_Val));
    }
   
    code = Code.concat(left_code, right_code, code);
    expr.set(type);
    expr.set(code);
    info.put(node, expr);
  }

  public void outATermSimpleExpr(ATermSimpleExpr node){
    info.put(node, info.get(node.getTerm()));
  }

  public void outAUnoptermSimpleExpr(AUnoptermSimpleExpr node){
    Information2 expr = new Information2();
    Information2 term = (Information2)info.get(node.getTerm());
    Codeptr term_code = term.getCode();
    Type term_type = term.getType();

    Codeptr code = null;
    if(node.getUnop() instanceof APlusUnop){
      code = term_code;
    }else{
      code = Code.concat(term_code, 
                          Code.gen(Opecode.o_ng, retType(term_type), Null_Val, Null_Val));
    }
    expr.set(term_type);
    expr.set(code);
    info.put(node, expr);         
  }

  public void outASimpleExprlowOptermSimpleExpr(ASimpleExprlowOptermSimpleExpr node){
    Information2 expr = new Information2();
    Information2 simple_expr = (Information2)info.get(node.getSimpleExpr());
    Codeptr simple_expr_code = simple_expr.getCode();
    Type simple_expr_type = simple_expr.getType();

    Information2 term = (Information2)info.get(node.getTerm());
    Codeptr term_code = term.getCode();
    Type term_type = term.getType();

    Type type = null;
    Codeptr code = null;
    if(simple_expr_type == type_int && term_type == type_int){
      type = type_int;
      code = Code.nullcode();
    }else{
      type=type_real;
      if(simple_expr_type == type_int){
        code = Code.gen(Opecode.o_flo);
      }else if(term_type == type_int){
        code = Code.gen(Opecode.o_flt);
      }else{
        code = c.nullcode();
      }
    }
    
    if(node.getLowOp() instanceof APlusLowOp){
      code = Code.concat(code, 
                          Code.gen(Opecode.o_ad, retType(term_type), Null_Val, Null_Val));
    }else{
      code = Code.concat(code,
                          Code.gen(Opecode.o_sb, retType(term_type), Null_Val, Null_Val)); 
    }
    code = Code.concat(simple_expr_code, term_code, code);
    expr.set(type);
    expr.set(code);
    info.put(node, expr);
  }

  public void outASimpleExprorOptermSimpleExpr(ASimpleExprorOptermSimpleExpr node){
    Information2 expr = new Information2();
    Information2 simple_expr = (Information2)info.get(node.getSimpleExpr());
    Codeptr simple_expr_code = simple_expr.getCode();

    Information2 term = (Information2)info.get(node.getTerm());
    Codeptr term_code = term.getCode();
    Codeptr code = Code.concat(simple_expr_code, term_code, 
                                Code.gen(Opecode.o_ior));
    expr.set(type_bool);
    expr.set(code);
    info.put(node, expr);
  }

  public void outAFactTerm(AFactTerm node){
    info.put(node, info.get(node.getFact()));
  }

  public void outATermhighOpfactTerm(ATermhighOpfactTerm node){
    Information2 high_op = new Information2();
    Information2 term = (Information2)info.get(node.getTerm());
    Codeptr term_code = term.getCode();
    Type term_type = term.getType();

    Information2 fact = (Information2)info.get(node.getFact());
    Codeptr fact_code = fact.getCode();
    Type fact_type = fact.getType();

    Type type = null;
    Codeptr code = null;
    if(node.getHighOp() instanceof ADivHighOp){
      type = type_int;
      code = Code.gen(Opecode.o_dv, Codetype.c_integer, Null_Val, Null_Val);
    }else if(node.getHighOp() instanceof AMultHighOp){ //|Z(*)
      if(term_type == type_int && fact_type == type_int){
        type = type_int;
        code = Code.gen(Opecode.o_mp, Codetype.c_integer, Null_Val, Null_Val);
      }else{
        type = type_real;
        if(term_type == type_int){
          code = Code.concat(Code.gen(Opecode.o_flo), 
                              Code.gen(Opecode.o_mp, Codetype.c_real, Null_Val, Null_Val));
        }else if(fact_type == type_int){
          code = Code.concat(Code.gen(Opecode.o_flt), 
                              Code.gen(Opecode.o_mp, Codetype.c_real, Null_Val, Null_Val));
        }else{
          code = Code.gen(Opecode.o_mp, Codetype.c_real, Null_Val, Null_Val);
        }
      }
    }else if(node.getHighOp() instanceof ASlashHighOp){
      type = type_real;

      Codeptr flo_code = null;
      Codeptr flt_code = null;
      if(term_type == type_int){
        flo_code = Code.gen(Opecode.o_flo);
      }else{
        flo_code = c.nullcode();
      }
      if(fact_type == type_int){
        flt_code = Code.gen(Opecode.o_flt);
      }else{
        flt_code = c.nullcode();
      }
      code = Code.concat(flt_code,  flo_code,
                       Code.gen(Opecode.o_dv, Codetype.c_real, Null_Val, Null_Val));
     }else{ // mod
      type = type_int;
      code = Code.gen(Opecode.o_mod, null_val, Null_Val, Null_Val);
    }
    code=Code.concat(term_code, fact_code, code);
    high_op.set(type);
    high_op.set(code);
    info.put(node, high_op);
  }

  public void outATermandOpfactTerm(ATermandOpfactTerm node){
    Information2 and_op = new Information2();
    Information2 term = (Information2)info.get(node.getTerm());
    Codeptr term_code = term.getCode();

    Information2 fact = (Information2)info.get(node.getFact());
    Codeptr fact_code = fact.getCode();

    Codeptr code=Code.concat(term_code, fact_code, 
                              Code.gen(Opecode.o_and));
    and_op.set(type_bool);
    and_op.set(code);
    info.put(node, and_op);
  }

  public void outAUnsignedConstantFact(AUnsignedConstantFact node){
    info.put(node, info.get(node.getUnsignedConstant()));
  }

  public void outAIdentFact(AIdentFact node){
    Information2 ident = (Information2)info.get(node.getIdent());
    info.put(node, codeIdent(ident.getString(), table.getLevel()));
  }

  public void outAVariableTailFact(AVariableTailFact node){
    Information2 ident = (Information2)info.get(node.getIdent());
    ConcreteAggregate variable_tail 
             = ((Information2)info.get(node.getVariableTail())).getList();
    Information2 fact = vtailCodeR(ident.getString(), variable_tail, table.getLevel());
    info.put(node, fact);
  }

  public void outAFuncCallWithParamFact(AFuncCallWithParamFact node){
    info.put(node, info.get(node.getFuncCallWithParam()));
  }

  public void outALparexprrparFact(ALparexprrparFact node){
    info.put(node, info.get(node.getExpr()));
  }

  public void outANotfactFact(ANotfactFact node){
    Information2 not_fact = new Information2();
    Information2 fact = (Information2)info.get(node.getFact());
    Codeptr fact_code = fact.getCode();
    Codeptr code = Code.concat(fact_code,Code.gen(Opecode.o_not));
    not_fact.set(type_bool);
    not_fact.set(code);
    info.put(node, not_fact);
  }

  public void outABuiltinFuncCallFuncCallWithParam(ABuiltinFuncCallFuncCallWithParam node){
    info.put(node, info.get(node.getBuiltinFuncCall()));
  }

  public void outAIdentFuncCallWithParam(AIdentFuncCallWithParam node){
    String ident = ((Information2)info.get(node.getIdent())).getString();
    ConcreteAggregate param = ((Information2)info.get(node.getActualParamList())).getList();
    Information2 func_call = funcwParamCode(ident, param, table.getLevel());
    info.put(node, func_call);
  }

  public void outASinBuiltinFuncCall(ASinBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Information2 expr=(Information2)info.get(node.getExpr());
    Codeptr expr_code=expr.getCode();
    Type expr_type=expr.getType();

    Codeptr code2 = null;
      if(expr_type==type_real){
        code2 = c.nullcode();
      }else{
        code2 = Code.gen(Opecode.o_flt);
      }

      Codeptr code=Code.concat(expr_code, code2, 
                                Code.gen(Opecode.o_csp, Builtinfunc.sin, Null_Val, Null_Val));
      func_call.set(type_real);
      func_call.set(code);
      info.put(node, func_call);
  }

  public void outACosBuiltinFuncCall(ACosBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Information2 expr=(Information2)info.get(node.getExpr());
    Codeptr expr_code=expr.getCode();
    Type expr_type=expr.getType();

    Codeptr code2 = null;
    if(expr_type==type_real){
      code2 = c.nullcode();
    }else{
      code2 = Code.gen(Opecode.o_flt);
    }

    Codeptr code=Code.concat(expr_code, code2, 
                              Code.gen(Opecode.o_csp, Builtinfunc.cos, Null_Val, Null_Val));
    func_call.set(type_real);
    func_call.set(code);
    info.put(node, func_call);
  }

  public void outAOddBuiltinFuncCall(AOddBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Codeptr expr_code = ((Information2)info.get(node.getExpr())).getCode();
    Codeptr code = Code.concat(expr_code,Code.gen(Opecode.o_odd));
    func_call.set(type_bool);
    func_call.set(code);
    info.put(node, func_call);
  }

  public void outAEvenBuiltinFuncCall(AEvenBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Codeptr expr_code = ((Information2)info.get(node.getExpr())).getCode();
    Codeptr code = Code.concat(expr_code,
                                Code.gen(Opecode.o_odd),
                                Code.gen(Opecode.o_not));
    func_call.set(type_bool);
    func_call.set(code);
    info.put(node, func_call);
  }

  public void outASqrBuiltinFuncCall(ASqrBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Information2 expr = (Information2)info.get(node.getExpr());
    Codeptr expr_code = expr.getCode();
    Type expr_type = expr.getType();
    Codeptr code3 = null;
    Codeptr code = Code.concat(expr_code, 
                               Code.gen(Opecode.o_sq, retType(expr_type), Null_Val, Null_Val));
    func_call.set(expr_type);
    func_call.set(code);
    info.put(node, func_call);
  }

  public void outAFArctanBuiltinFuncCall(AFArctanBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Information2 expr = (Information2)info.get(node.getExpr());
    Codeptr expr_code = expr.getCode();
    Type expr_type = expr.getType();
    Codeptr code2 = null;
    if(expr_type == type_real){
      code2 = c.nullcode();
    }else{
      code2 = Code.gen(Opecode.o_flt);
    }
    Codeptr code = Code.concat(expr_code, code2, 
                                Code.gen(Opecode.o_csp, Builtinfunc.atn, Null_Val, Null_Val));
    func_call.set(type_real);
    func_call.set(code);
    info.put(node, func_call);
  }

  public void outAFSqrtBuiltinFuncCall(AFSqrtBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Information2 expr=(Information2)info.get(node.getExpr());
    Codeptr expr_code=expr.getCode();
    Type expr_type=expr.getType();
    Codeptr code2 = null;
    if(expr_type == type_real){
      code2 = c.nullcode();
    }else{
      code2 = Code.gen(Opecode.o_flt);
    }
    Codeptr code = Code.concat(expr_code, code2, 
                                Code.gen(Opecode.o_csp, Builtinfunc.sqt, Null_Val, Null_Val));
    func_call.set(type_real);
    func_call.set(code);
    info.put(node, func_call);
  }

  public void outAFAbsBuiltinFuncCall(AFAbsBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Information2 expr = (Information2)info.get(node.getExpr());
    Codeptr expr_code = expr.getCode();
    Type expr_type = expr.getType();
    Codeptr code = Code.concat(expr_code, 
                                Code.gen(Opecode.o_ab, retType(expr_type), Null_Val, Null_Val));
    func_call.set(expr_type);
    func_call.set(code);
    info.put(node, func_call);
  }

  public void outAFRoundBuiltinFuncCall(AFRoundBuiltinFuncCall node){
    Information2 func_call = new Information2();
    func_call.set(type_real);
    func_call.set(Code.nullcode());
    info.put(node, func_call);
  }

  public void outAFTruncBuiltinFuncCall(AFTruncBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Codeptr expr_code = ((Information2)info.get(node.getExpr())).getCode();
    Codeptr code = Code.concat(expr_code,Code.gen(Opecode.o_trc));
    func_call.set(type_int);
    func_call.set(code);
    info.put(node, func_call);
  }

  public void outAFExpBuiltinFuncCall(AFExpBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Information2 expr = (Information2)info.get(node.getExpr());
    Codeptr expr_code = expr.getCode();
    Type expr_type = expr.getType();
    Codeptr code2 = null;
    if(expr_type == type_real){
      code2 = c.nullcode();
    }else{
      code2 = Code.gen(Opecode.o_flt);
    }
    Codeptr code = Code.concat(expr_code, code2, 
                               Code.gen(Opecode.o_csp, Builtinfunc.exp, Null_Val, Null_Val));
    func_call.set(type_real);
    func_call.set(code);
    info.put(node, func_call);
  }


  public void outAFLnBuiltinFuncCall(AFLnBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Information2 expr = (Information2)info.get(node.getExpr());
    Codeptr expr_code = expr.getCode();
    Type expr_type = expr.getType();
    Codeptr code2 = null;
    if(expr_type == type_real){
      code2 = c.nullcode();
    }else{
      code2 = Code.gen(Opecode.o_flt);
    }
    Codeptr code = Code.concat(expr_code, code2, 
                                Code.gen(Opecode.o_csp, Builtinfunc.log, Null_Val, Null_Val));
    func_call.set(type_real);
    func_call.set(code);
    info.put(node, func_call);
  }

  public void outAFChrBuiltinFuncCall(AFChrBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Codeptr expr_code = ((Information2)info.get(node.getExpr())).getCode();
    Codeptr code = Code.concat(expr_code,Code.gen(Opecode.o_chr));
    func_call.set(type_char);
    func_call.set(code);
    info.put(node, func_call);
  }

  public void outAFOrdBuiltinFuncCall(AFExpBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Information2 expr = (Information2)info.get(node.getExpr());
    Codeptr expr_code = expr.getCode();
    Type expr_type = expr.getType();
    Codeptr code = Code.concat(expr_code, 
                                Code.gen(Opecode.o_ord, retType(expr_type), Null_Val, Null_Val));
    func_call.set(type_int);
    func_call.set(code);
    info.put(node, func_call);
  }

  public void outAFEofBuiltinFuncCall(AFEofBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Codeptr code = Code.concat(Code.gen(Opecode.o_lao, null_val, 5.0, Null_Val),
                               Code.gen(Opecode.o_eof));
    func_call.set(type_bool);
    func_call.set(code);
    info.put(node, func_call);
  }

  public void outAFEolnBuiltinFuncCall(AFEolnBuiltinFuncCall node){
    Information2 func_call = new Information2();
    Codeptr code = Code.concat(Code.gen(Opecode.o_lao, null_val, 5.0, Null_Val), 
                               Code.gen(Opecode.o_csp, Builtinfunc.eoln , Null_Val, Null_Val));
    func_call.set(type_bool);
    func_call.set(code);
    info.put(node, func_call);
  }

  public void outAActualParamList(AActualParamList node){
    info.put(node, info.get(node.getArgList()));
  }

  public void outAArgListArgList(AArgListArgList node){
    Information2 arg_list = (Information2)info.get(node.getArgList());
    ConcreteAggregate list = arg_list.getList();
    list.add(info.get(node.getExpr()));
    info.put(node, arg_list);
  }

  public void outAExprArgList(AExprArgList node){
    Information2 arg_list = new Information2();
    ConcreteAggregate list =  new ConcreteAggregate();
    list.add(info.get(node.getExpr()));
    arg_list.set(list);
    info.put(node, arg_list);
  }

  public void outASubscriptListVariableTail(ASubscriptListVariableTail node){
    Information2 subscript_list = (Information2)info.get(node.getSubscriptList());
    ConcreteAggregate list = subscript_list.getList();
    if(node.getCustomTypeTail() instanceof AVariableTailCustomTypeTail){
      ConcreteAggregate list2 
                      = ((Information2)info.get(node.getCustomTypeTail())).getList();
      Iterator it =  list2.iterator();
      while(it.hasNext()){
        list.add(it.next());
      }
    }
    info.put(node, subscript_list);
  }

  public void outAFieldIdentVariableTail(AFieldIdentVariableTail node){
    Information2 var_tail = new Information2();
    ConcreteAggregate list = new ConcreteAggregate();
    list.add(info.get(node.getIdent()));
    if(node.getCustomTypeTail() instanceof AVariableTailCustomTypeTail){
      ConcreteAggregate list2 
           = ((Information2)info.get(node.getCustomTypeTail())).getList();
      Iterator it =  list2.iterator();
      while(it.hasNext()){
        list.add(it.next());
      }
    }
    var_tail.set(list);
    info.put(node, var_tail);
  }

  public void outAVariableTailCustomTypeTail(AVariableTailCustomTypeTail node){
    info.put(node, info.get(node.getVariableTail()));
  }

  public void outASubscriptListSubscriptList(ASubscriptListSubscriptList node){
    Information2 subscript_list = (Information2)info.get(node.getSubscriptList());
    ConcreteAggregate list =  subscript_list.getList();
    list.add(info.get(node.getExpr()));
    info.put(node, subscript_list);
  }

  public void outAExprSubscriptList(AExprSubscriptList node){
    Information2 sub_list = new Information2();
    ConcreteAggregate list =  new ConcreteAggregate();
    list.add(info.get(node.getExpr()));
    sub_list.set(list);
    info.put(node, sub_list);
  }

  public void outANumberConstConstant(ANumberConstConstant node){
    info.put(node, info.get(node.getNumberConst()));
  }

  public void outAIdentConstant(AIdentConstant node){
    try{
      Information2 ident_constant = new Information2();
      Constant ident = (Constant)table.lookup(((Information2)info.get(node.getIdent())).getString());
      ident_constant.set(ident.getValue());
      ident_constant.set(ident.getType());
      info.put(node, ident_constant);
    }catch(TableError e){
      System.out.println(e.getMessage());
    }
  }

  public void outAUnopConstant(AUnopConstant node){
    try{
      Information2 unop_const = new Information2();
      Constant ident  = (Constant)table.lookup(((Information2)info.get(node.getIdent())).getString());
      Type type= ident.getType();
      String value = ident.getValue();
      if(node.getUnop() instanceof AMinusUnop){
        value= "-" + value;
      }
      unop_const.set(value);
      unop_const.set(type);
      info.put(node, unop_const);
    }catch(TableError e){
      System.out.println(e.getMessage());
    }
  } 

  public void outACharConstConstant(ACharConstConstant node){
    info.put(node, info.get(node.getCharConst()));
  }

  public void outAUnsignedNumberNumberConst(AUnsignedNumberNumberConst node){
    info.put(node, info.get(node.getUnsignedNumber()));
  }

  public void outAUnopNumberConst(AUnopNumberConst node){
    Information2 unsigned_number = (Information2)info.get(node.getUnsignedNumber());
    if(node.getUnop() instanceof AMinusUnop){
      String number = "-"+unsigned_number.getString();
      unsigned_number.set(number); 
    }
    info.put(node, unsigned_number);
  }

  public void outAUnsignedNumber(AUnsignedNumber node){
    info.put(node, info.get(node.getNumber()));
  }

  public void outARealNumber(ARealNumber node){
    Information2 real_num = new Information2();
    real_num.set(node.getReal().getText().trim());
    real_num.set(type_real);
    info.put(node, real_num);
  }

  public void outAIntegerNumber(AIntegerNumber node){
    Information2 int_num = new Information2();
    int_num.set(node.getInteger().getText().trim());
    int_num.set(type_int);
    info.put(node, int_num);
  }

  public void outACharConst(ACharConst node){
    info.put(node, info.get(node.getCharacter()));
  }

  public void outACharacter(ACharacter node){
    Information2 charact = new Information2();
    String char_const = node.getChar().getText().trim();
    Type type=null;
    if(char_const.length() == 3){
      type=type_char;
    }else{
      type=type_string;
    }
    charact.set(char_const);
    charact.set(type);
    info.put(node, charact);
  }

  public void outAUnsignedNumberUnsignedConstant(AUnsignedNumberUnsignedConstant node){
   Information2 unsigned_const = new Information2();
   Information2 unsigned_number = (Information2)info.get(node.getUnsignedNumber());
   double number = Double.parseDouble(unsigned_number.getString());
   Type type = unsigned_number.getType();
   Codeptr code = Code.gen(Opecode.o_ldc, retType(type), number, Null_Val);
   unsigned_const.set(type);
   unsigned_const.set(code);
   info.put(node, unsigned_const);
  }

  public void outACharConstUnsignedConstant(ACharConstUnsignedConstant node){
    Information2 unsigned_const = new Information2();
    String char_const = ((Information2)info.get(node.getCharConst())).getString();
    Codeptr code = Code.nullcode();
    if(char_const.length() == 3){
      unsigned_const.set(type_char);
      code = Code.gen(Opecode.o_ldc, Codetype.c_character, c.setCharString(char_const), Null_Val);
    }else{
      int str_len = char_const.length() - 2;
      code = Code.concat(Code.gen(Opecode.o_lca, (int)c.setCharString(char_const), Null_Val, Null_Val),
                         Code.gen(Opecode.o_ldc, Codetype.c_integer, str_len, Null_Val),
                         Code.gen(Opecode.o_ldc, Codetype.c_integer, str_len, Null_Val));
      unsigned_const.set(type_string);
    }
    unsigned_const.set(code);
    info.put(node, unsigned_const);
  }

  public void outABTrueUnsignedConstant(ABTrueUnsignedConstant node){
    Information2 unsigned_const = new Information2();
    unsigned_const.set(type_bool);
    unsigned_const.set(Code.gen(Opecode.o_ldc, Codetype.c_boolean, 1.0, Null_Val));
    info.put(node, unsigned_const);
  }

  public void outABFalseUnsignedConstant(ABFalseUnsignedConstant node){
    Information2 unsigned_const = new Information2();
    unsigned_const.set(type_bool);
    unsigned_const.set(Code.gen(Opecode.o_ldc, Codetype.c_boolean, 0.0, Null_Val));
    info.put(node, unsigned_const);
  }

  public int retType(Type type){
    if(type==type_int) return Codetype.c_integer;
    else if(type == type_real) return Codetype.c_real;
    else if(type == type_char) return Codetype.c_character;
    else if(type ==type_bool) return Codetype.c_boolean;
    else  return Codetype.c_etc;
  }

  public int addVal1(int val, hasParameter proc){
    int rval = proc.getVal1();
    int val12 = rval + val;
    proc.setVal1(val12);
    proc.setVal2(val12);
    if(proc instanceof Program){
      return 9 + rval; //offset̏l
    }else{
      return 5 + rval;
    }
  }

  public Codeptr retCode(hasParameter proc){
    if(proc instanceof Function){
       Type type=((Function)proc).getType();
       return Code.gen(Opecode.o_ret, retType(type), Null_Val, Null_Val);
    }else{
       return Code.gen(Opecode.o_ret, Codetype.c_etc, Null_Val, Null_Val);
    }
  }

  public Codeptr changeCode(Codeptr code){
    int inst_f=code.getHead().getInst().getF();
    int inst_l=code.getHead().getInst().getL();
    double inst_a1=code.getHead().getInst().getA1();
    double inst_a2=code.getHead().getInst().getA2();
    Codeptr return_code=null;
    if(inst_f == Opecode.o_lod){
      return_code=Code.gen(Opecode.o_lda, inst_l, inst_a1, inst_a2);
    }else if(inst_f==Opecode.o_ldo){
      return_code=Code.gen(Opecode.o_lao, inst_l, inst_a1, inst_a2);
    }else{
      return_code=code;
    }
    return return_code;
  }

  public Codeptr cutCode(Codeptr code){
    if(code.getHead().getNext() != null){
      if(code.getHead().getNext().getInst().getF() == Opecode.o_chk){
        code.setTail(code.getHead());
      }
    }
    return code;
  }

  public Information2 funcwParamCode(String name, ConcreteAggregate actual_param_list, int level){
    Information2 ret = new Information2();
    try{
      hasParameter proc = null;
      Type type = null;
      Codeptr code = Code.nullcode();
      Entry entry = table.lookup(name);
      if(entry instanceof Procedure){
        proc = (Procedure)entry;
      }else if(entry instanceof Function){
        proc = (Function)entry;
        type = ((Function)entry).getType();
      }
      int i = proc.getNumberOfParameter();
      Codeptr param_code = getParamCode(proc.getParameter(), actual_param_list);
      code= Code.concat(Code.gen(Opecode.o_mst, null_val, (double)(level - entry.getLevel()), Null_Val),
                         param_code, 
                         Code.gen(Opecode.o_cup, i, (double)proc.getLabel(), Null_Val));
      ret.set(type);
      ret.set(code);
   }catch(TableError e){
      System.out.println(e.getMessage());
    }
    return ret;
  }

  public Codeptr getParamCode(Iterator f_it, ConcreteAggregate actual_param_list){
    Codeptr code=c.nullcode();
    Iterator a_it = actual_param_list.iterator();
    while(a_it.hasNext()){
      Information2 expr = (Information2)a_it.next();
      Type expr_type = expr.getType();
      Codeptr expr_code = expr.getCode();
      Parameter formal_param = (Parameter)f_it.next();
      Codeptr code1=null, code2=null;
      if(formal_param instanceof ValueParameter){
        code1 = changeCode(expr_code);
      }else{
        code1 = expr_code;
      }
      if(expr_type == type_bool){
       code2=Code.gen(Opecode.o_chk, Codetype.c_boolean, Null_Val, 1.0);
      }else{
       code2=c.nullcode();
      }
      int check=expr_code.getHead().getInst().getF();
      code = Code.concat(code, code1, code2);
    }
    return code;
  }

  public Information2 codeIdent(String name, int level){
    Information2 ret = new Information2();
    try{
      Codeptr code = Code.nullcode();
      Entry entry = table.lookup(name);
      Type type=((hasType)entry).getType();
      if(entry instanceof Constant){
        double val=0.0;
        if(type==type_char){
          val = c.setCharString(((Constant)entry).getValue());
        }else{
          val = Double.parseDouble(((Constant)entry).getValue());
        }
        code= Code.gen(Opecode.o_ldc, retType(type), val, Null_Val);
      }else if(entry instanceof Variable){
        int addr = ((Variable)entry).getAddress();
        int vlevel = entry.getLevel();
        if(vlevel == 1){
          code = Code.gen(Opecode.o_ldo, retType(type), addr,  Null_Val);
        }else{
	    code= Code.gen(Opecode.o_lod, retType(type), level - vlevel, addr);
        }
      }else if(entry instanceof Function){
        int label = ((Function)entry).getLabel();
        int flevel = entry.getLevel();
        code= Code.concat(Code.gen(Opecode.o_mst, null_val, level - flevel, Null_Val),
                         Code.gen(Opecode.o_cup, null_val, label,  Null_Val));
      }else if(entry instanceof Parameter){
        int addr = ((Parameter)entry).getAddress();
        if((Parameter)entry instanceof VariableParameter){
          code=Code.concat(Code.gen(Opecode.o_lod, Codetype.c_addr, Null_Val, addr),
                             Code.gen(Opecode.o_ind, retType(type), Null_Val, Null_Val));  
        }else{
          code= Code.gen(Opecode.o_lod, retType(type), Null_Val, addr);
        }
      }
      if(code == null){
        code = Code.nullcode();
      }
      ret.set(type);
      ret.set(code);
    }catch(TableError e){
      System.out.println(e.getMessage());
    }
    return ret;
  }

  public Codeptr vtailCodeL(String name, ConcreteAggregate variable_tail, Codeptr expr_code, Type expr_type, int level){
    try{
      Entry entry=table.lookup(name);
      Type type=((hasType)entry).getType();

      Codeptr code = c.nullcode();
      boolean flag = false;
      int inc = 0; 

      Iterator tail_it = variable_tail.iterator();

      while(type instanceof ArrayType || type instanceof RecordType){
        if(type instanceof ArrayType){
          ConcreteAggregate index_list = ((ArrayType)type).getIndex();
          Iterator index_it = index_list.iterator();        
        
          if(!flag){
	    if(entry.getLevel()==1){
	      code = Code.gen(Opecode.o_lao, null_val, ((hasAddress)entry).getAddress()+inc, Null_Val);
            }else if(entry instanceof VariableParameter){
              code=Code.gen(Opecode.o_lod, Codetype.c_addr,
	         level - entry.getLevel(), ((Parameter)entry).getAddress()+inc);
	    }else{
	      code=Code.gen(Opecode.o_lda, null_val,
		  level - entry.getLevel(), ((hasAddress)entry).getAddress()+inc);
            }
          }

          while(index_it.hasNext()){
            Information2 expr  = (Information2)tail_it.next();
            Codeptr code2 = expr.getCode();
            Index index=(Index)index_it.next();
            int start = index.getLow();
            int end = index.getHigh();
            int address = index.getAddress();

            if(expr.getType() != type_int){ 
              code2 = Code.concat(cutCode(code2),
                                   Code.gen(Opecode.o_ord, retType(expr.getType()), Null_Val, Null_Val));
            }

            Codeptr code4=null;
            if(start == 0){ 
              code4 = c.nullcode();
            }else{ 
              code4 = Code.gen(Opecode.o_dec, Codetype.c_integer, start,  Null_Val);
            }

            code=Code.concat(code, 
                              code2, 
                              Code.gen(Opecode.o_chk,Codetype.c_integer, start, end),
                              code4, 
                              Code.gen(Opecode.o_ixa, Codetype.c_integer, address, Null_Val));
          }

          if(!tail_it.hasNext()){
            type = ((ArrayType)type).getType();
            Codeptr code3 = null;
            if(type == type_bool){
              code3 = Code.gen(Opecode.o_chk, Codetype.c_boolean, 0.0, 1.0);
            }else if(type == type_char){
              code3 = Code.gen(Opecode.o_chk, Codetype.c_character, 0.0, 63.0);
            }else if(type == type_real &&  expr_type == type_int){
              code3 = Code.gen(Opecode.o_flt);
            }else{
             code3 = Code.nullcode();
            }
            return Code.concat(code, 
                                expr_code, 
                                code3, 
                                Code.gen(Opecode.o_sto, retType(type), Null_Val, Null_Val));
	  }
          flag=true;
          inc = 0;
          type=((ArrayType)type).getType();
        }else if(type instanceof RecordType){
          Information2 field_name = (Information2)tail_it.next();
          Field field = (Field)table.lookup(field_name.getString(), ((RecordType)type).getField());
          type = field.getType();
          inc = inc + field.getAddress();

          if(!tail_it.hasNext()){
            Codeptr code3_4 = null;
            if(type == type_bool){
              code3_4 = Code.gen(Opecode.o_chk, Codetype.c_boolean, 0.0, 1.0);
            }else if(type == type_char){
              code3_4 = Code.gen(Opecode.o_chk, Codetype.c_character, 0.0, 63.0);
            }else if(type == type_real &&  expr_type == type_int){
              code3_4 = Code.gen(Opecode.o_flt);
            }else{
              code3_4 = Code.nullcode();
            }

            if(flag){
              Codeptr code2 = null;
              if(inc == 0){
                code2 = Code.nullcode();
              }else{
                code2 = Code.gen(Opecode.o_inc, Codetype.c_addr, inc, Null_Val);
              }
              code = Code.concat(code, code2, expr_code, code3_4,
                                 Code.gen(Opecode.o_sto, retType(type), Null_Val, Null_Val));
              return code;
            }

            Codeptr code4 = null;
            if(entry.getLevel() == 1){
              code4=Code.gen(Opecode.o_sro, retType(type),  ((hasAddress)entry).getAddress()+inc, Null_Val);
            }else{
              code4=Code.gen(Opecode.o_str, retType(type), 
                            (double)(level - field.getLevel()), ((hasAddress)entry).getAddress()+inc);
            }
            code = Code.concat(code, expr_code, code3_4, code4);
            return code;
          }else{ //܂Ȃ
            if(type instanceof ArrayType){
              if(flag){
                Codeptr code2=null;
                if(inc==0){
                  code2=c.nullcode();
                }else{
	          code2=Code.gen(Opecode.o_inc, Codetype.c_addr, field.getAddress()+inc, Null_Val);
                }
                code=Code.concat(code,code2);
                inc = 0;
              }
	    }
	  }
	}
      }
      return code;
    }catch(TableError e){
      System.out.println(e.getMessage());
    }
    return Code.nullcode();
  }

  public Information2 vtailCodeR(String name, ConcreteAggregate variable_tail, int level){
    Information2 ret = new Information2();
    try{
      Entry entry=table.lookup(name);
      Type type=((hasType)entry).getType();

      Codeptr code = c.nullcode();
      boolean flag = false;
      int inc = 0; 

      Iterator tail_it = variable_tail.iterator();

      while(type instanceof ArrayType || type instanceof RecordType){
        if(type instanceof ArrayType){
          ConcreteAggregate index_list = ((ArrayType)type).getIndex();
          Iterator index_it = index_list.iterator();        
        
          if(!flag){
	    if(entry.getLevel()==1){
	      code = Code.gen(Opecode.o_lao, null_val, ((hasAddress)entry).getAddress()+inc, Null_Val);
            }else if(entry instanceof VariableParameter){
              code=Code.gen(Opecode.o_lod, Codetype.c_addr,
	         level - entry.getLevel(), ((Parameter)entry).getAddress()+inc);
	    }else{
	      code=Code.gen(Opecode.o_lda, null_val,
		  level - entry.getLevel(), ((hasAddress)entry).getAddress()+inc);
            }
	  }

          while(index_it.hasNext()){
            Information2 expr  = (Information2)tail_it.next();
            Codeptr expr_code = expr.getCode();
            Index index=(Index)index_it.next();
            int start = index.getLow();
            int end = index.getHigh();
            int address = index.getAddress();

            Codeptr code2 = null;
            if(expr.getType() != type_int){
              code2 = Code.concat(cutCode(expr_code),
                                  Code.gen(Opecode.o_ord, retType(expr.getType()), Null_Val, Null_Val));
            }else{
              code2 = expr_code;
            }

            Codeptr code4=null;
            if(start == 0){ 
              code4 = c.nullcode();
            }else{
              code4 = Code.gen(Opecode.o_dec, Codetype.c_integer, start,  Null_Val);
            }

            code=Code.concat(code, 
                              code2, 
                              Code.gen(Opecode.o_chk,Codetype.c_integer, start, end),
                              code4, 
                              Code.gen(Opecode.o_ixa, Codetype.c_integer, address, Null_Val));
	  }

          if(!tail_it.hasNext()){
            type = ((ArrayType)type).getType();
            code = Code.concat(code, Code.gen(Opecode.o_ind, retType(type), Null_Val, Null_Val));
	    ret.set(type);
	    ret.set(code);
	    return ret;
	  }
          flag=true;
          inc = 0;
          type=((ArrayType)type).getType();
        }else if(type instanceof RecordType){ //R[hȂ
          Information2 field_name = (Information2)tail_it.next();
          Field field = (Field)table.lookup(field_name.getString(), ((RecordType)type).getField());
          type = field.getType();
          inc = inc + field.getAddress();

          if(!tail_it.hasNext()){
            if(flag){
              code=Code.concat(code, Code.gen(Opecode.o_ind, retType(type), inc, Null_Val));
	      ret.set(type);
	      ret.set(code);
	      return ret;
            }

            Codeptr code2=null;
            if(entry.getLevel() == 1){
              code2=Code.gen(Opecode.o_ldo, retType(type), ((hasAddress)entry).getAddress()+inc, Null_Val);
            }else{
              code2=Code.gen(Opecode.o_lod, retType(type), 
                            (double)(level - field.getLevel()), ((hasAddress)entry).getAddress()+inc);
            }
            code=Code.concat(code, code2);
	    ret.set(type);
	    ret.set(code);
	    return ret;
          }else{
            if(type instanceof ArrayType){
              if(flag){
                Codeptr code2=null;
                if(inc==0){
                  code2=c.nullcode();
                }else{
	          code2=Code.gen(Opecode.o_inc, Codetype.c_addr, field.getAddress()+inc, Null_Val);
                }
                code=Code.concat(code,code2);
                inc = 0;
              }
	    }
          }
	}
      }
      ret.set(type);
      ret.set(code);
    }catch(TableError e){
      System.out.println(e.getMessage());
    }
    return ret;
  }
}


